#lang racket
;; Now we can implementing a multiplex "queue"
;; in terms of other queues.
;; [Beware that the multiplexes one is not actually
;;  a queue anymore, because it doesn't preserve
;;  order, but maybe we don't care.]

(define-struct q (en-evt de-evt))

(define (create-queue)
  (define enqueue-ch (make-channel))
  (define dequeue-ch (make-channel))
  
  (define (serve items)
    (sync (handle-evt enqueue-ch
                      (lambda (item)
                        (serve (append items
                                       (list item)))))
          (if (empty? items)
              never-evt
              (handle-evt dequeue-ch
                          (lambda (reply-ch)
                            (channel-put reply-ch (first items))
                            (serve (rest items)))))))
  
  (thread (lambda () (serve empty)))
  
  (make-q (lambda (v)
            (channel-put-evt enqueue-ch v))
          (lambda (reply-ch)
            (channel-put-evt dequeue-ch reply-ch))))

(define q1 (create-queue))
(define q2 (create-queue))

(define (multiplex-q q1 q2)
  (make-q (lambda (v)
            (choice-evt ((q-en-evt q1) v)
                        ((q-en-evt q2) v)))
          (lambda (reply-ch)
            (choice-evt ((q-de-evt q1) reply-ch)
                        ((q-de-evt q2) reply-ch)))))

(define q1+q2 (multiplex-q q1 q2))

(sync ((q-en-evt q1+q2) 10))
(sync ((q-en-evt q1+q2) 20))

(define (dequeue q)
  (define reply-ch (make-channel))
  (sync ((q-de-evt q) reply-ch))
  (channel-get reply-ch))

(dequeue q1+q2)
(dequeue q1+q2)
(thread (lambda () (sleep 2) (sync ((q-en-evt q1+q2) 30))))
(dequeue q1+q2)
