#lang racket

;; A queue that accepts a predicate before supplying an
;; item from the queue. The server now needs to keep
;; a list of requests, and the synchronization point is
;; the request response, not the request acceptance.

(define-struct q (en-evt de-evt))

(define (create-queue)
  (define enqueue-ch (make-channel))
  (define dequeue-ch (make-channel))
  
  ;; A req (cons pred reply-ch)
  
  (define (serve items reqs)
    (sync 
     (handle-evt enqueue-ch
                 (lambda (item)
                   (serve (append items
                                  (list item))
                          reqs)))
     (handle-evt dequeue-ch
                 (lambda (req)
                   (serve items (cons req reqs))))
     (apply
      choice-evt
      (map (lambda (req)
             (match req
               [(cons pred reply-ch)
                (define ok-items (filter pred items))
                (if (empty? ok-items)
                    never-evt
                    (handle-evt (channel-put-evt reply-ch (first ok-items))
                                (lambda (_)
                                  (serve (remq (first ok-items) items)
                                         (remq req reqs)))))]))
           reqs))))

  (thread (lambda () (serve empty empty)))
  
  (make-q (lambda (v)
            (channel-put-evt enqueue-ch v))
          (lambda (pred reply-ch)
            (guard-evt
             (lambda ()
               (channel-put dequeue-ch (cons pred reply-ch))
               reply-ch)))))

(define q1 (create-queue))

(sync ((q-en-evt q1) 10))
(sync ((q-en-evt q1) 21))

(define (dequeue q pred)
  (define reply-ch (make-channel))
  (sync ((q-de-evt q) pred reply-ch)))

(dequeue q1 odd?)
(dequeue q1 even?)
(thread (lambda () 
          (sleep 1) 
          (sync ((q-en-evt q1) 30))
          (sleep 1) 
          (sync ((q-en-evt q1) 41))))
(dequeue q1 odd?)
